#include <msp/core/refptr.h>
#include <msp/io/file.h>
#include <msp/strings/format.h>
#include "bmploader.h"
#include "imageloader.h"
#ifdef WITH_LIBPNG
#include "png/pngloader.h"
#endif
#ifdef WITH_LIBJPEG
#include "jpeg/jpegloader.h"
#endif
#ifdef WITH_DEVIL
#include "devil/devilloader.h"
#endif
#ifdef WITH_QUARTZ
#include "quartz/quartzloader.h"
#endif

using namespace std;

namespace Msp {
namespace Graphics {

bool ImageLoader::registered_loaders_changed = false;

ImageLoader::ImageLoader():
	source(0)
{ }

ImageLoader::~ImageLoader()
{
	delete source;
}

ImageLoader *ImageLoader::open_file(const string &fn)
{
	try
	{
		RefPtr<IO::BufferedFile> file = new IO::BufferedFile(fn);
		ImageLoader *loader = open_io(*file);
		loader->source = file.release();
		return loader;
	}
	catch(const unsupported_image_format &e)
	{
		throw unsupported_image_format(format("%s: %s", fn, e.what()));
	}
}

ImageLoader *ImageLoader::open_io(IO::Seekable &io)
{
	(void)RegisteredImageLoader<BmpLoader>::reg;
#ifdef WITH_LIBPNG
	(void)RegisteredImageLoader<PngLoader>::reg;
#endif
#ifdef WITH_LIBJPEG
	(void)RegisteredImageLoader<JpegLoader>::reg;
#endif
#ifdef WITH_DEVIL
	(void)RegisteredImageLoader<DevilLoader>::reg;
#endif
#ifdef WITH_QUARTZ
	(void)RegisteredImageLoader<QuartzLoader>::reg;
#endif

	list<RegisterBase *> &loaders = get_registered_loaders();
	if(registered_loaders_changed)
	{
		registered_loaders_changed = false;
		loaders.sort(signature_size_compare);
	}

	if(loaders.empty())
		throw unsupported_image_format("no loaders");

	vector<char> sig_buf(loaders.back()->get_signature_size());
	unsigned sig_len = io.read(&sig_buf[0], sig_buf.size());
	string signature(sig_buf.begin(), sig_buf.end());

	ImageLoader *loader = 0;
	for(list<RegisterBase *>::const_iterator i=loaders.begin(); (!loader && i!=loaders.end()); ++i)
		if((*i)->detect(signature))
			loader = (*i)->create(io);

	io.seek(0, IO::S_BEG);

	if(!loader)
	{
		string sig_hex;
		for(unsigned i=0; i<sig_len; ++i)
		{
			if(i)
				sig_hex += ' ';
			sig_hex += format("%02X", static_cast<unsigned char>(sig_buf[i]));
		}
		throw unsupported_image_format(sig_hex);
	}

	return loader;
}


ImageLoader::RegisterBase::RegisterBase()
{
	get_registered_loaders().push_back(this);
	registered_loaders_changed = true;
}


list<ImageLoader::RegisterBase *> &ImageLoader::get_registered_loaders()
{
	static list<RegisterBase *> regs;
	return regs;
}

bool ImageLoader::signature_size_compare(RegisterBase *r1, RegisterBase *r2)
{
	return r1->get_signature_size()<r2->get_signature_size();
}

} // namespace Graphics
} // namespace Msp
